# 1 € generative Art
# creates generative art (a drawing of a continuous line, stored as tuples of coordinates (x,y,draw))
# coordinate values range from 0 to 1
# sends coordinates ("x*10000,y*10000") via serial to Arduino when s/S is pressed
# quits when q/Q is pressed
# published under the beer ware license
# written by Niklas Roy in 2024

import pygame
import sys
import math
import time
import serial
import serial.tools.list_ports
import random
import numpy as np
from generate_lines import generatePoints, stretch, soft_step, sin_osc, numberOfAlgorithms
from plotterfont import plot_text
from datetime import datetime
import requests
import hashlib
import os


# ------------------------------------------------------------------------------ INIT
pygame.init()               # Initialize Pygame
pygame.mixer.init()         # Initialize the audio mixer
clock = pygame.time.Clock() # Initialize the clock

FPS = 30                    # program loop fps
useSerial = 1;              # 0= do not use serial port
changeTime = 100            # algorithm changes every _(eg 100) seconds

#............................ SCREEN

                            # screen sizes for built in laptop with second screen
squareSize = 741            # square drawing area (size and position) on the screen
squX = 83                   # can be temporarily changed with WASD and ER for resizing
squY = 175


WHITE = (255, 255, 255)     # Define colors
GREY = (128,128 ,128)
BLACK = (0, 0, 0)
RED   = (255, 0, 0)
infoFont = pygame.font.Font("fonts/Futura LT Bold.ttf",195) #Futura Bold BT.ttf", 130) # Define font and size

#............................ SOUND
coinSound = pygame.mixer.Sound("sounds/coin.wav")   # Load WAV file
confirmSound = pygame.mixer.Sound("sounds/confirm.wav")   # Load WAV file
soundInput = 0              #sound
sloSound = 0

#............................ GENERAL
maxPoints = 1000            # maximum amount of points for generated graphics
selectNewest = 0            #flag to select the newest algorithm
plottedLineIndex = -1       # number of line(/point) which is currently being plotted
penX = 0                    # pen position
penY = 0
running = True
millisOffset = 0            # timing offset as plotting takes time
plotStartTime = 0
plotEndTime   = 0
marqueeTimeOffset = 0
pBlink = 0                  # blink state in previous cycle
percent = 0
pPercent = 0                # previous percent of the finished drawing

#............................ LINE ALGOS
random.seed(int(time.time() * 1000))           # leet seed
pLineSelector  = 0
lineSelectorA  =  round(random.random()*(numberOfAlgorithms-3))
lineSelectorB  =  lineSelectorA + round(random.random()*(numberOfAlgorithms-3)) +1
lineSelectorB  = lineSelectorB % (numberOfAlgorithms+1)
lastSelectors = []
for i in range(20): # memorize last selectors so they don't repeat too fast
    lastSelectors.append(lineSelectorA)  # Add lineSelectorA to the list
    lastSelectors.append(lineSelectorB)

switchFade = 0
pSwitchFade = 0
softness = 100 # camera smoothness
softSize = 1   # camera zoom
softX = 0.5 # camera pan
softY = 0.5 # camera tilt
pPoints = []
softPX = [0.5] * maxPoints # soften the point positions
softPY = [0.5] * maxPoints
soften = 0#15 filter point positions
pLength = 0 # previous length of the path
pMaxLength = 0 # previous maximum length of longest line

#............................ STATE MACHINE
STATE       = 1             # current state
NORMAL      = 1
PLOTTING    = 2
FINISHING   = 3
PLOTTED     = 4
ERROR       = 0

# ------------------------------------------------------------------------------ CREATE WINDOW

# window = pygame.display.set_mode((0, 0), pygame.FULLSCREEN)   # Get the display surface and set the window to fullscreen
# window = pygame.display.set_mode((600, 600))                  # Get the display surface and set the window
display_info = pygame.display.Info()                            # Get the display info for the second monitor (index 1)
display_count = pygame.display.get_num_displays()
if display_count > 1:                                           # Check if the second display exists
    display_rect = pygame.display.get_desktop_sizes()[1]        # Get the second display's information
    window = pygame.display.set_mode((1680, 1050), pygame.FULLSCREEN, display=1) # Create the window on the second display
else:
    print("Second display not found. Using primary display.")
    window = pygame.display.set_mode((1680, 1050), pygame.FULLSCREEN)
                                # overwrite screen sizes
    squareSize = 741            # square drawing area (size and position) on the screen
    squX = 81                   # can be temporarily changed with WASD and ER for resizing
    squY = 52

pygame.display.set_caption("1€ Generative Art")
pygame.mouse.set_visible(False)                                 # Hide the mouse pointer

print ("")
print ("+-------------------------------+")
print ("| Generative Art 1€ says hello! |")
print ("+-------------------------------+")
print ("|       hit q/Q to quit         |")
print ("+-------------------------------+")
print ("")

#--------------------------------------------------------------------------- DATA LOGGING

def load_secret_key(file_path):
    try:
        if not os.path.exists(file_path): # Check if file exists
            print(f"Secret key file not found: {file_path}")
            return None
        with open(file_path, 'r') as key_file: # Read and strip the key
            secret_key = key_file.read().strip()
        if not secret_key: # Validate key is not empty
            print("Secret key is empty")
            return None
        return secret_key
    except Exception as e:
        print(f"Error loading secret key: {e}")
        return None

def log(logText):
    # Local logging
    now = datetime.now()
    date_time_string = now.strftime("%Y-%m-%d %H:%M:%S")
    print(date_time_string+" : "+logText)

    logfile = 'log.dat'
    try:
        with open(logfile, 'a') as file:
            file.write(date_time_string+" | "+logText + '\n')
    except Exception as e:
        print(f"An error occurred while writing to the logfile: {e}")

    # Web logging (optional)
    web_log_url = "" #TODO: URL for logging goes haere <--

    # Load secret key from external file
    try:
        with open('pw.dat', 'r') as key_file:
            secret_key = key_file.read().strip()
    except Exception as e:
        print(f"Error loading secret key: {e}")
        return

    if web_log_url and secret_key:
        try:
            # Create signature exactly like PHP script (MD5)
            signature = hashlib.md5((secret_key + logText).encode()).hexdigest()

            # Prepare payload
            payload = {
                'message': logText,
                'timestamp': date_time_string,
                'signature': signature
            }

            # Send log to web server
            response = requests.post(web_log_url, data=payload, timeout=5)

            # Optional: print error if web logging fails
            if response.status_code != 200:
                print(f"Web logging failed: {response.status_code}")
        except Exception as e:
            print(f"Error in web logging: {e}")

# Log program start with script name
log("////////////////////////////////////// PROGRAM STARTED // "+sys.argv[0])

#--------------------------------------------------------------------------- MILLISECONDS
def getMillis():
    global millisOffset
    if STATE == NORMAL:
        millisOffset *= .993 # reduce the offset
        millisOffset -= 1
        millisOffset = int(millisOffset)
        if millisOffset<1:
            millisOffset = 0
    return int(time.time() * 1000) - int(millisOffset)

#--------------------------------------------------------------------------- MILLISECONDS for MARQUEE
def getMarqueeMillis():
    return int(time.time() * 1000)

#--------------------------------------------------------------------------- SERIAL INIT

def processPointsfindArduinoPort():
    """Find the first available serial port"""
    ports = list(serial.tools.list_ports.comports())
    return ports[0].device if ports else None

def connectArduino(port='COM4', baudrate=115200):
    """Establish serial connection with Arduino"""
    try:
        ser = serial.Serial(port, baudrate, timeout=1)
        time.sleep(2)  # Allow Arduino to reset
        print(f"Connected to {port}")
        return ser
    except Exception as e:
        print(f"Connection failed: {e}")
        return None

def serialSend(message):
    """
    Send a message to Arduino with special handling for 'p' messages

    Args:
        message (str): Message to send to Arduino
    """
    if ser is None:
        print("Error: Serial connection not established.")
        return
    full_message = message + '\n'

    # Special handling for messages starting with 'p'
    if message.startswith('p'):

        # Send the message
        ser.write(full_message.encode('utf-8'))
        print(f"Sent: {message}")

        while True:
            # Wait for and check the response
            response = serialReceive()
            if response and response.startswith('p'):
                # Received confirmation, can proceed
                break
    else:
        # Normal send for other messages
        ser.write(full_message.encode('utf-8'))
        if not message.startswith('l'): # print the sent messages, except blinking lights to keep it less verbose
            print(f"Sent: {message}")

def serialReceive():
    """
    Receive a message from Arduino

    Returns:
        str: Received message or None
    """
    if ser is None:
        print("Error: Serial connection not established.")
        return
    if ser.in_waiting > 0:
        message = ser.readline().decode('utf-8').strip()
        print(f"Received: {message}")
        return message
    return None

ser = connectArduino()
serialSend("l1,0,0") # lights off
serialSend("l2,0,0") #
serialSend("l3,0,0") #

#--------------------------------------------------------------------------- SEND SQUARE
def sendSquare():
    print("Drawing frame")
    serialSend(f"f0,0,0")

#--------------------------------------------------------------------------- CHECK FOR ERROR
def checkError(confirmationLetter, errorTimeOut):
    global STATE
    serialResponse = ""
    loopStartTime = time.time()  # Record the start time for the loop

    while time.time() - loopStartTime < errorTimeOut:  # Wait for confirmation or timeout
        pygame.event.pump()  # Avoid OS thinking the program is stalled

        if serialResponse == confirmationLetter:  # Success condition
            return

        if serialResponse == "Z":  # Error condition
            STATE = ERROR
            log('ARDUINO ERROR: ' + confirmationLetter)
            return

        serialResponse = serialReceive()  # Update serialResponse

    # If the loop exits, it must be due to timeout
    STATE = ERROR
    log('PYTHON TIMEOUT ERROR: ' + confirmationLetter)
    return

#--------------------------------------------------------------------------- KEYBOARD
def checkKeyboard():
    global running, switchFade, pSwitchFade, selectNewest,  squX, squY, squareSize, STATE
    for event in pygame.event.get():
        if event.type == pygame.QUIT:       # QUIT:
            ser.close()                     # Close the serial connection
            running = False
        elif event.type == pygame.KEYDOWN:
            if event.key == pygame.K_f:     # ----> F: draw FRAME
                if STATE == NORMAL:
                    sendSquare()
            if event.key == pygame.K_p:     # ----> P: save PNG
                print ('save png')
                saveDrawingAsPNG()
            if event.key == pygame.K_q:     # ----> Q: QUIT
                print ('QUIT!')
                running = False
            if event.key == pygame.K_1:     # ----> 1: COIN inserted
                if STATE == NORMAL:
                    STATE = PLOTTING
                    log('TEST PLOT VIA KEYBOARD')
                    marqueeTimeOffset = getMarqueeMillis()
            if event.key == pygame.K_c:     # ----> C: CUT
                if STATE == NORMAL:
                    serialSend("c0,0,0")
                    checkError("C",30)
            if event.key == pygame.K_RETURN:# ----> RETURN: switch to latest generative algorithm
                print("Return pressed")
                selectNewest=1
                pSwitchFade = switchFade+1

            if event.key == pygame.K_w:     # ----> WASD: MOVE screen content
                squX = squX + 1
                print (f"squX:{squX} squY:{squY} squareSize:{squareSize}")
            if event.key == pygame.K_a:
                squY = squY + 1
                print (f"squX:{squX} squY:{squY} squareSize:{squareSize}")
            if event.key == pygame.K_s:
                squX = squX - 1
                print (f"squX:{squX} squY:{squY} squareSize:{squareSize}")
            if event.key == pygame.K_d:
                squY = squY - 1
                print (f"squX:{squX} squY:{squY} squareSize:{squareSize}")

            if event.key == pygame.K_r:     # ----> RE: RESIZE screen content
                squareSize = squareSize +1
                print (f"squX:{squX} squY:{squY} squareSize:{squareSize}")
            if event.key == pygame.K_e:
                squareSize = squareSize -1
                print (f"squX:{squX} squY:{squY} squareSize:{squareSize}")

            if event.key == pygame.K_UP:    # ----> UP ARROW_ MOVE paper UP
                print("Up arrow pressed")
                serialSend("u0,0,0")
            if event.key == pygame.K_DOWN:  # ----> DOWN ARROW_ MOVE paper DOWN
                print("Down arrow pressed")
                serialSend("d0,0,0")
#--------------------------------------------------------------------------- DISTANCE
def dist(x1, y1, x2, y2):
    return math.sqrt((x2 - x1)**2 + (y2 - y1)**2)
#--------------------------------------------------------------------------- PLOT AN ARTWORK
def plot(pointsToSend):
    global maxPoints, milliseconds, plottedLineIndex, penX, penY, percent, pPercent
    penX  = pointsToSend[0][0]
    penY  = pointsToSend[0][1]
    updateLineDrawing(1)
    updateEuroInfo("0,00€",RED,WHITE)
    pygame.display.flip()

    xSend = 0
    ySend = 0
    pXSend = 0
    pYSend = 0

    serialSend(f"p{int(crop(pointsToSend[0][0],0,1)*10000)},{int(crop(pointsToSend[0][1],0,1)*10000)},0") # send first coordinate

    freq = 0
    formatted_string = ""
    for i in range(maxPoints):  # iterate over all coordinates
        plottedLineIndex = i
        xSend = int(crop(pointsToSend[i][0],0,1)*10000)
        ySend = int(crop(pointsToSend[i][1],0,1)*10000)
        penX  = pointsToSend[i][0]
        penY  = pointsToSend[i][1]
        if (abs(xSend-pXSend)>20 or abs(ySend-pYSend)>10 ): # only send coordinates, if pen will move more than half a millimeter
            serialSend(f"p{xSend},{ySend},1") # send coordinates

            dur = dist(xSend,ySend,pXSend,pYSend)/6
            if (i%2==0):
                freq = pow(abs(xSend-5000)/75,2)
            else:
                freq = pow(abs(ySend-5000)/75,2)

            if (dur>20 or i%8==0):
                if (i%32==0):
                    freq/=2
                elif (i%32==1):
                    freq*=2
                generateBeep(dur, freq, .95)

            pXSend = xSend
            pYSend = ySend

        updateLineDrawing(1)
        pPercent = percent;
        percent = math.ceil(i/10)

        eurocents = (percent) * 0.01 # Convert percent to Eurocents (multiply by 0.01)
        formatted_string = f"{eurocents:0.2f}€ " # Format the Eurocents value to 2 decimal places and add the '€' symbol
        formatted_string = formatted_string.replace('.', ',') # Replace the dot with a comma (European style formatting)
        updateEuroInfo(formatted_string,RED,WHITE) # write progress number 0...100 in info box

        if (percent != pPercent):
            generateBeep(20, 1000+(percent-pPercent)*500, 1) # beep when counting money

        pygame.display.flip()
    updateLineDrawing(0)
    generateBeep(150, freq, .5)
    updateEuroInfo(formatted_string,RED,WHITE) # write progress number 0...100 in info box
    pygame.display.flip()

#--------------------------------------------------------------------------- DRAW CROSSHAIR PEN
def drawPen():#
    global squX, squY, squareSize
    crossSize=10
    startPoint = (penY*squareSize+squX,penX*squareSize+squY-crossSize)
    endPoint =  (penY*squareSize+squX,penX*squareSize+squY+crossSize)
    drawThickAALine(window, BLACK, startPoint, endPoint, 2)
    startPoint = (penY*squareSize+squX+crossSize,penX*squareSize+squY)
    endPoint =  (penY*squareSize+squX-crossSize,penX*squareSize+squY)
    drawThickAALine(window, BLACK, startPoint, endPoint, 2)

#--------------------------------------------------------------------------- DRAW LINES
def drawLines(pointArray):  # attention! screen is built in reversed!

    for i in range(maxPoints-1):                    # iterate over point array
        x1 = pointArray[i][1]*squareSize+squX
        x2 = pointArray[i+1][1]*squareSize+squX
        y1 = pointArray[i][0]*squareSize+squY
        y2 = pointArray[i+1][0]*squareSize+squY
        startPoint = (x1,y1)
        endPoint = (x2,y2)

        if (STATE == NORMAL or STATE == PLOTTING):
            if (i>=plottedLineIndex):                   # draw the line in black (virtual) or red (when already plotted)
                drawThickAALine(window, BLACK, startPoint, endPoint, 2)
            else:
                drawThickAALine(window, RED, startPoint, endPoint, 2)
        else:
                drawThickAALine(window, WHITE, startPoint, endPoint, 2)

#--------------------------------------------------------------------------- DRAW ONE THICK ANTIALIASED LINE
def drawThickAALine(surface, color, start, end, width):
    # Calculate the direction vector of the line
    dx = end[0] - start[0]
    dy = end[1] - start[1]

    # Normalize the direction vector
    length = (dx ** 2 + dy ** 2) ** 0.5
    if (length == 0): length = .0001
    dx /= length
    dy /= length

    # Draw several thin lines around the main line to simulate thickness
    for i in range(-width//2, width//2 + 1):
        offset_x = i * dy
        offset_y = -i * dx
        start_offset = (start[0] + offset_x, start[1] + offset_y)
        end_offset = (end[0] + offset_x, end[1] + offset_y)
        pygame.draw.aaline(surface, color, start_offset, end_offset)

#--------------------------------------------------------------------------- CREATE TEXT BITMAP FOR INFO BOX
def createTextBitmap(text, fontColor):
    rendered_text = infoFont.render(text, True, fontColor)
    mirrored_text = pygame.transform.flip(rendered_text, True, False) # Flip the text horizontally
    return pygame.transform.rotate(mirrored_text, -90)        # Rotate the text 90 degrees

#--------------------------------------------------------------------------- UPDATE GRAPHICS (DRAW EVERYTHING ON SCREEN)
def updateLineDrawing(showPen):

    global dataY, softPoints, milliseconds, clock, lineSelectorA, lineSelectorB, name1, name2, fade, softX, softY, softSize, zoom
    global totalLength, lengthDifference, maxLengthDifference, maxLength, squX, squY, squareSize

    window.fill(BLACK)                                                          # clear BGND (BLACK)
    if STATE == NORMAL or STATE == PLOTTING:
        pygame.draw.rect(window, (255,255,255), pygame.Rect(squX-20, squY-20, squareSize+40, squareSize+40)) # WHITE square BGND for line drawing
    else:
        pygame.draw.rect(window, (255,0,0), pygame.Rect(squX-20, squY-20, squareSize+40, squareSize+40)) # RED square BGND for line drawing

    drawLines(softPoints)       # draw all lines
    if (showPen == 1):          # draw pen
        drawPen()
#--------------------------------------------------------------------------- UPDATE INFO BOX
def updateInfoBox(infoText,fontColor,bgndColor):
    infoX = squX+squareSize+80  # info rect size
    infoY = squY-20
    infoHeight = 250
    infoWidth = squareSize+40
    pygame.draw.rect(window, bgndColor, pygame.Rect(infoX, infoY, infoHeight, infoWidth)) # rect BGND for info

    # write in info box
    text = createTextBitmap(infoText,fontColor)
    text_width, text_height = text.get_size()
    window.blit(text, (infoX +infoHeight/2 - text_width/2 , infoY + infoWidth/2 - text_height/2))

def updateEuroInfo(infoText, fontColor, bgndColor):
    infoX = squX + squareSize + 80  # Info rect position
    infoY = squY - 20
    infoHeight = 250
    infoWidth = squareSize + 40
    # Draw the background rectangle for the info box
    pygame.draw.rect(window, bgndColor, pygame.Rect(infoX, infoY, infoHeight, infoWidth))

    # Define the starting position for the first character
    char_x = (infoX + infoHeight / 2 - 128)  # vertical position on real screen
    char_y = 720  # horizontal on real screen

    # Render and position each character
    for char in infoText:
        char_surface = createTextBitmap(char, fontColor)  # Create surface for the character
        char_width, char_height = char_surface.get_size()

        # Adjust positions for specific characters
        char_y = char_y  # Start with the default y-position
        if char in ['4', '5']:  # Slightly adjust problematic characters
            char_y += 0
        elif char == ',':
            char_y += 75
        elif char == '€':
            char_y -= 0

        # Blit the character at the current position
        window.blit(char_surface, (char_x, char_y))
        # Increment x position for the next character
        char_y -= 95*1.5


#--------------------------------------------------------------------------- SAVE PNG
def saveDrawingAsPNG():
    filename = f"drawings/dr_{milliseconds}.png"    # Convert millisecondsto a string and create a filename
    if window is not None: # Check if the drawing window surface exists
        pygame.image.save(window, filename) # Save the current window surface as a PNG file
        print(f'Drawing saved as {filename}')
    else:
        print('No drawing surface found!')

#--------------------------------------------------------------------------- CROP VALUE
def crop(value, lower_bound, upper_bound):
    if (value<lower_bound):
        return lower_bound
    if (value>upper_bound):
        return upper_bound
    return value

#--------------------------------------------------------------------------- PROCESS POINTS
def processPoints(milliseconds):
    """Process points, handle algorithm switching and point generation."""
    global lineSelectorA, lineSelectorB, selectNewest, pSwitchFade, softSize, softX, softY, pLength, pMaxLength, softPoints

    seconds = (milliseconds/1000)
    switchInterval = seconds/changeTime
    switchFade = switchInterval+.5 - round(switchInterval) # 0...1

    if (switchFade<pSwitchFade):                    # switch algos
        random.seed(milliseconds)                   # select random algorithms based on ms
        lineSelectorB = lineSelectorA
        lineSelectorA += random.randint(1,numberOfAlgorithms-1)
        lineSelectorA = lineSelectorA % (numberOfAlgorithms+1)

        testSelector = 1                            # test if this selector has been used recently, if so take the next one
        while testSelector == 1:
            if lineSelectorA in lastSelectors:
                lineSelectorA = (lineSelectorA+1) % (numberOfAlgorithms+1)
            else:
                testSelector = 0
                lastSelectors.pop() # Remove the last value by popping
                lastSelectors.insert(0, lineSelectorA) # Insert lineSelectorA at the beginning (index 0)


        if (selectNewest == 1):
            lineSelectorA = numberOfAlgorithms
            selectNewest=0 # reset flag
        print("selectors:", lineSelectorB, " -> ", lineSelectorA)
    pSwitchFade = switchFade
    switchInterval = round(switchInterval)

    # generate two sets of points which will be then mixed / faded
    points1, name1 = generatePoints(lineSelectorA, milliseconds/10, maxPoints)
    points2, name2 = generatePoints(lineSelectorB, milliseconds/10, maxPoints)
    points = []

    #calculate waves for depth wiggle effect
    wiggleX = math.sin(milliseconds/12345)/6
    wiggleY = math.sin(milliseconds/23456)/6

    # fade/morph between two sets of points
    fade = switchFade  #0...1
    fade = (fade + soft_step(fade*5)/5)/2
    fade2 = (fade*2)-.5 #-.5 ... 1.5

    for i in range(maxPoints):
        fade3 = fade2+(i/maxPoints)-.5 #0... 0.5
        fade3 = crop(fade3, 0,1)
        f = 1-(math.cos(fade3*math.pi)/2+.5)
        f1 = 1-f
        # wiggle
        depth = (i/maxPoints)-.4
        x = points1[i][0]*f+points2[i][0]*f1+wiggleX*depth
        y = points1[i][1]*f+points2[i][1]*f1+wiggleY*depth
        points.append((x,y,1))

    # Find the coordinate maxima:
    minX = float('inf')     # Initialize min and max values for x and y
    maxX = float('-inf')
    minY = float('inf')
    maxY = float('-inf')
    # Find minX, maxX, minY, and maxY
    for point in points:
        x, y, _ = point  # Unpack the coordinates
        minX = min(minX, x)
        maxX = max(maxX, x)
        minY = min(minY, y)
        maxY = max(maxY, y)

    # prepare camera pan and zoom
    softness = 100
    xSize = maxX-minX
    ySize = maxY-minY
    size  = max(xSize,ySize)
    softSize = (softSize*softness+size)/(softness+1)
    xCenter = (maxX+minX)/2
    yCenter = (maxY+minY)/2
    softX = (softX*softness+xCenter)/(softness+1)
    softY = (softY*softness+yCenter)/(softness+1)
    shiftedPoints = []
    zoom = (math.sin(milliseconds/8456)/10)+.9 # sinewave zoom effect
    i=0
    zoomOsc = (math.sin(milliseconds/6351)+1)/2 # depth oscillation
    for point in points:
        x, y, _ = point  # Unpack the coordinates
        newX = x-softX
        newY = y-softY
        depth = 1-(i/maxPoints) #zoom
        depth= ((depth/5)+.7) * zoomOsc + (1-zoomOsc)*.7
        newX = newX/(softSize)*zoom*depth
        newY = newY/(softSize)*zoom*depth
        i += 1
        newX += .5  #shift to center
        newY += .5
        shiftedPoints.append((newX,newY,1))

    # soften the points
    softPoints = []
    for i in range(maxPoints):
        px = crop((softPX[i]*soften+shiftedPoints[i][0])/(soften+1),.005,.995)
        py = crop((softPY[i]*soften+shiftedPoints[i][1])/(soften+1),.005,.995)
        softPX[i] = px
        softPY[i] = py
        softPoints.append((softPX[i],softPY[i],1))

    #calculate the length of the path
    totalLength = 0
    maxLength = 0
    for i in range(maxPoints-1):
        a1, b1, d = softPoints[i]
        a2, b2, d = softPoints[i+1]
        length = math.sqrt(pow(a1-a2,2)+pow(b1-b2,2))
        totalLength += length
        maxLength = max(length, maxLength) # find the longest single path

    return totalLength, maxLength
#--------------------------------------------------------------------------- GENERATE SOUND
def generateSound(totalLength, pLength, maxLength, pMaxLength):
    """Generate sound based on path length changes."""
    global sloSound
    lengthDifference = totalLength - pLength

    soundInput = abs(lengthDifference)          # Set the duration and frequency of the "tic" sound
    sloSound = (soundInput+sloSound*20)/21
    soundOutput = crop(soundInput-sloSound-.01,0,100)
    duration_ms =  soundOutput*3                # milliseconds
    frequency =    soundInput*1000              # Hz
    sample_rate = 44100                         # Samples per second
    num_samples = int(duration_ms * sample_rate / 1000) # Generate samples for the sine wave
    samples = (np.sin(2 * np.pi * frequency * np.arange(num_samples) / sample_rate) * 32767).astype(np.int16)
    tic_sound = pygame.mixer.Sound(samples)     # Create a Sound object with the sine wave samples
    if (STATE != ERROR):
        tic_sound.set_volume(.5)
        tic_sound.play()                            # Play the "tic" sound

    return lengthDifference

def generateBeep(duration_ms, frequency, volume=1.0, sample_rate=44100, ramp_ms=5):
    # Volume adjustment for different frequencies
    volumeMultiplier = crop(900-frequency, 40, 900)/900
    volume = volume * 0.95 * volumeMultiplier

    try:
        num_samples = int(duration_ms * sample_rate / 1000)
        ramp_samples = int(ramp_ms * sample_rate / 1000)

        # Generate basic sine wave
        t = np.arange(num_samples)
        sine_wave = np.sin(2 * np.pi * frequency * t / sample_rate)

        # Create envelope
        envelope = np.ones(num_samples)

        # Attack ramp (linear from 0 to 1)
        if ramp_samples > 0:
            envelope[:ramp_samples] = np.linspace(0, 1, ramp_samples)
            # Release ramp (linear from 1 to 0)
            envelope[-ramp_samples:] = np.linspace(1, 0, ramp_samples)

        # Apply envelope and convert to 16-bit audio
        samples = (sine_wave * envelope * 32767 * volume).astype(np.int16)

        sound = pygame.mixer.Sound(samples)
        sound.play()

    except Exception as e:
        print(f"Error generating sound: {e}")

#--------------------------------------------------------------------------- write in plotterfont
def writeLine(string,scale, space, mirrored):
    global squX, squY, squareSize
    vectorList = []
    #vectorList.extend(plot_text(string, squX - 20, squY - 30, scale, space, mirrored))
    vectorList.extend(plot_text(string, squX - 20,  squareSize+squY+50, scale, space, mirrored))
    for i in range(len(vectorList)):
        start, end = vectorList[i]
        pygame.draw.line(window, GREY, start, end)

############################################################################ MAIN LOOP

while running:
    P_STATE = STATE
    checkKeyboard()
    message = serialReceive() # Check for and process incoming messages

    if (message == "COIN"):   # received a real coin!
        if (STATE == NORMAL):
            STATE = PLOTTING
            log("*** GOT COIN! ***")
        else:
            log(f"GOT COIN // BUT CURRENTLY IN STATE: {STATE}")

    if (message == "OUT OF PAPER"):
        log ('ERROR: OUT OF PAPER!')
        STATE = ERROR


    #..................................................... ERROR STATE
    if (STATE == ERROR):
        milliseconds = getMillis()
        # Process points and get path length
        totalLength, maxLength = processPoints(milliseconds)
        # Update global length tracking
        lengthDifference = generateSound(totalLength, pLength, maxLength, pMaxLength)
        pLength = totalLength
        maxLengthDifference = abs(maxLength - pMaxLength)
        pMaxLength = maxLength
        updateLineDrawing(0) # draw everything on the screen (except the pen)

        text_array = ["ERROR", ""]
        infoText = text_array[math.floor((getMarqueeMillis()-marqueeTimeOffset)/1000)%len(text_array)]

        updateInfoBox(infoText,WHITE,RED)
        blink    = math.floor((getMarqueeMillis()-marqueeTimeOffset)/1000)%2
        if (pBlink != blink):
            serialSend("l1,0,0") # lights off
            serialSend("l2,0,0")
            serialSend("l3,0,0")
        pBlink = blink

        pygame.display.flip()
    #..................................................... NORMAL STATE
    if (STATE == NORMAL):
        if (P_STATE != STATE): # turn lights off
            serialSend("l1,0,0")
            serialSend("l3,0,0")
        milliseconds = getMillis()
        # Process points and get path length
        totalLength, maxLength = processPoints(milliseconds)
        # Update global length tracking
        lengthDifference = generateSound(totalLength, pLength, maxLength, pMaxLength)
        pLength = totalLength
        maxLengthDifference = abs(maxLength - pMaxLength)
        pMaxLength = maxLength
        updateLineDrawing(0) # draw everything on the screen (except the pen)

        #writeLine(f"SEED:{milliseconds} OFF:{millisOffset} - A:{lineSelectorA:03} - B:{lineSelectorB:03}", 8,15, 1)

        text_array = ["", "BUY", "ART", "NOW", "", "", "INSERT", "COIN", ""]
        infoText = text_array[math.floor((getMarqueeMillis()-marqueeTimeOffset)/1500)%len(text_array)] # marquee text


        if (infoText == "INSERT" or infoText == "COIN"):
            blink   = math.floor((getMarqueeMillis()-marqueeTimeOffset)/750)%2 # blinking arrow
        else:
            blink = 1

        if (pBlink!=blink):
            serialSend(f"l2,{blink},0") # blink arrow off
        pBlink = blink

        updateInfoBox(infoText,RED,WHITE)
        pygame.display.flip()

    #..................................................... GOT COIN
    if (STATE == PLOTTING):

        plotStartTime = getMarqueeMillis()

        generateBeep(600, 880, 1.5)
        time.sleep(.4)
        generateBeep(100, 880*1.5, 1.5)
        time.sleep(1)

        serialSend("l1,0,0")
        serialSend("l2,0,0")
        serialSend("l3,1,0")

        updateLineDrawing(0) # draw everything on the screen (except the pen)
        updateEuroInfo("0,00€",RED,WHITE)
        pygame.display.flip()

        #generateBeep(duration_ms, frequency, volume=1.0, sample_rate=44100):

        plot(softPoints)                   # plot drawing -- this takes long!
        serialSend(f"w{milliseconds},0,0") # write seed

        print (f"print seed: {milliseconds}")
        while (serialReceive()!="W"):      # wait for confirmation
            pygame.event.pump()            # avoid that OS thinks the prog is stalled
        print (f"seed printed")

        serialSend(f"g{milliseconds},0,0") # move drawing into screen area
        while (serialReceive()!="G"):      # wait for confirmation
            pygame.event.pump()            # avoid that OS thinks the prog is stalled
        time.sleep(2)
        STATE = FINISHING
    #..................................................... Finish drawing with stamp and cutting
    if (STATE == FINISHING):

        #confirmSound.play()                 #confirmation sound
        generateBeep(300, 660, .9)

        plottedLineIndex = -1              # set plotted line back to -1
        loopStartTime = time.time()           # Record the start time for he loop
        loopDuration = 2.5                    # Duration in seconds
        while time.time() - loopStartTime < loopDuration: # loop for 2.5 s
            updateLineDrawing(0)
            updateInfoBox("",WHITE,BLACK)
            pygame.display.flip()

        marqueeTimeOffset = getMarqueeMillis()
        # . . . . . . . . . . . . . . . . . . . . . . . .  STAMP
        print("stamp")
        generateBeep(150, 990, .9)
        serialSend("e0,0,0")  # send 'end' signal for stamping
        serialResponse = ""  # Initialize serial response
        loopStartTime = time.time()
        confirmationLetter = "E"
        errorTimeOut = 20

        while ((serialResponse != confirmationLetter) and
               (serialResponse != "Z") and
               (time.time() - loopStartTime <= errorTimeOut)):
            # Update and process the UI and time-sensitive tasks
            updateLineDrawing(0)  # Draw everything on the screen (except the pen)
            updateInfoBox("STAMP", WHITE, RED)
            pygame.display.flip()
            serialResponse = serialReceive()  # Receive serial response

        # Check for success or error
        if serialResponse == "Z":
            STATE = ERROR
            log('ARDUINO ERROR: ' + str(confirmationLetter))
        if time.time() - loopStartTime > errorTimeOut:  # Timeout reached
            STATE = ERROR
            log('PYTHON TIMEOUT ERROR: ' + str(confirmationLetter))
        print("end of stamping sequence")

        # . . . . . . . . . . . . . . . . . . . . . . . .  CUT
        if (STATE != ERROR):
            print("cut")
            generateBeep(150, 1320, .9)
            serialSend("l1,1,0") # lamp 1 on (take out)
            serialSend("c0,0,0")  # send 'cut' signal for cutting
            serialResponse = ""  # Initialize serial response
            loopStartTime = time.time()
            confirmationLetter = "C"
            errorTimeOut = 40

            while ((serialResponse != confirmationLetter) and
                   (serialResponse != "Z") and
                   (time.time() - loopStartTime <= errorTimeOut)):

                    # Update and process the UI and time-sensitive tasks
                    updateLineDrawing(0)            # draw everything on the screen (except the pen)
                    updateInfoBox("CUT", WHITE, RED)
                    pygame.display.flip()
                    serialResponse = serialReceive()  # Receive serial response

            if serialResponse == "Z":
                STATE = ERROR
                log('ARDUINO ERROR: ' + str(confirmationLetter))

            if time.time() - loopStartTime > errorTimeOut:  # Timeout reached
                STATE = ERROR
                log('PYTHON TIMEOUT ERROR: ' + str(confirmationLetter))


            serialSend("l2,0,0")
            serialSend("l3,0,0")
            print("end of cutting sequence")

        # . . . . . . . . . . . . . . . . . . . . . . . .  DRAW NEW FRAME
        if (STATE != ERROR):
            generateBeep(150, 660, .9)
            sendSquare()
            print ("plot new frame")
            marqueeTimeOffset = getMarqueeMillis()
            while (serialReceive()!="F"):       # wait for confirmation
                updateLineDrawing(0)
                text_array = ["PLEASE", "WAIT"]
                infoText = text_array[math.floor((getMarqueeMillis()-marqueeTimeOffset)/1500)%len(text_array)]
                updateInfoBox(infoText, WHITE, RED)
                pygame.display.flip()          # flip the screen content
            print ("new frame is plotted")
            serialSend("l1,0,0")
            serialSend("l2,1,0")
            serialSend("l3,0,0")
            plotEndTime = getMarqueeMillis()         # calculate time offset - from now on animation starts again
            millisOffset += plotEndTime - plotStartTime
            plotStartTime = 0
            plotEndTime = 0

            print ("STATE = NORMAL")
            STATE = NORMAL
            marqueeTimeOffset = getMarqueeMillis()


#--------------------------------------------------------------------------- QUIT
ser.close()
pygame.quit()
sys.exit()
